{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {
    "collapsed": false
   },
   "source": [
    "# Laboratorio: Algoritmos de búsqueda\n",
    "\n",
    "## Grupo 2\n",
    "\n",
    "* Diego Fernandez García\n",
    "* Sol Roman \n",
    "* Javier de la Rosa Fernández\n",
    "* Aritz Beraza Garayalde"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
   ],
   "source": [
    "from copy import deepcopy\n",
    "import time\n",
    "import math\n",
    "\n",
    "# Estado inicial. Cada 'torre' de bloques se representa por un array donde el primer elemento es \n",
    "# la base de la torre y el último la cima\n",
    "\n",
    "MESA = [\"C\", \"B\"]\n",
    "TORRE = [\"E\", \"D\", \"A\"]"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "collapsed": false
   },
   "source": [
    "#### Clase Estado"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
   ],
   "source": [
    "# Cada estado guarda información de\n",
    "#  * Cajas en la torre (de abajo a arriba)\n",
    "#  * Cajas en la mesa\n",
    "#  * Movimientos hasta llegar a el.\n",
    "class Estado:\n",
    "    META = [\"E\", \"D\", \"C\", \"B\", \"A\"]\n",
    "    def __init__(self, t, m, p):\n",
    "        self.torre = t\n",
    "        self.mesa = m\n",
    "        self.solucion = p\n",
    "        # para A*\n",
    "        self.coste = math.inf # infinito por defecto\n",
    "\n",
    "    def isMeta(self):\n",
    "        return Estado.META == self.torre\n",
    "\n",
    "    # MoverCaja. Como una puede realizar máximo un movimiento, con señalar\n",
    "    # la caja a mover es suficiente para conocer el movimiento.\n",
    "    def moverCaja(self, caja, log=False):\n",
    "        # Si la caja se encuentra en la mesa, la añadimos a la torre\n",
    "        if caja in self.mesa:\n",
    "            self.torre.append(caja)\n",
    "            self.mesa.remove(caja)\n",
    "            if log:\n",
    "                print(\"Mover la caja [{}] de la mesa a la torre\".format(caja))\n",
    "        # Si la caja es la última en la torre, la ponemos en la mesa\n",
    "        elif caja == self.torre[-1]:\n",
    "            self.mesa.append(self.torre.pop())\n",
    "            if log:\n",
    "                print(\"Mover a caja [{}] de la torre a la mesa\".format(caja))\n",
    "\n",
    "        #self.estado[mov[1]].append(self.estado[mov[0]].pop())\n",
    "        self.solucion.append(caja)\n",
    "        return self\n",
    "\n",
    "    # Espandir Nodo para generar el siguiente nivel del árbol\n",
    "    def expandirNodo(self):\n",
    "        expandedNodes = []\n",
    "        # Para cada caja en la mesa, expandimos moviendo las cajas a la torre\n",
    "        for caja in self.mesa:\n",
    "            expandedNodes.append(deepcopy(self).moverCaja(caja))\n",
    "        # También se expande la última caja de la torre moviendola a la mesa\n",
    "        if len(self.torre) > 0:\n",
    "            expandedNodes.append(deepcopy(self).moverCaja(self.torre[-1]))\n",
    "        return expandedNodes\n",
    "    \n",
    "    def set_coste(self, coste):\n",
    "        self.coste = coste\n",
    "        return self\n",
    "\n",
    "    def print(self):\n",
    "        print(\"--Estado---------\")\n",
    "        print(\"TORRE :: {}\".format(self.torre))\n",
    "        print(\"MESA  :: {}\".format(self.mesa))\n",
    "\n",
    "\n",
    "# Tests:\n",
    "# Almacena info de 3 torres y una lista de acciones\n",
    "#e1 = Estado([\"E\", \"C\"],[\"B\", \"D\", \"A\"], [\"A\", \"E\"])\n",
    "#print(e1.torre)\n",
    "#print(e1.isMeta())\n",
    "# Si estado es meta, isMeta() devuelve True\n",
    "#e2=Estado([\"E\", \"D\", \"C\", \"B\", \"A\"],[],[])\n",
    "#e3=Estado([],[\"E\", \"D\", \"C\", \"B\", \"A\"],[])\n",
    "#print(e2.isMeta())\n",
    "#print(e3.isMeta())"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "collapsed": false
   },
   "source": [
    "#### Búsqueda en Amplitud"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
   ],
   "source": [
    "# Búsqueda de amplitud\n",
    "# Implementación del algoritmo de búsqueda de amplitud.\n",
    "# Iterativamente, partiendo del estado inicial realizamos:\n",
    "# A. Para cada uno de los estados posibles en un punto\n",
    "#   1. Es este estado el estado Meta? \n",
    "#        1a. SI: Parar ejecución, hemos llegado a la meta.\n",
    "#        1b. NO: Continuar ejecución\n",
    "#   2. Expandimos el nodo y añadimos a la lista de nodos del siguiente nivel\n",
    "#   3. Iterar el proceso para cada uno de los nodos del siguiente nivel\n",
    "# B. Tras encontrar una solución: Partiendo del estado inicial, reproducir\n",
    "#    la solución paso a paso y llegar a la meta.\n",
    "def busquedaAmplitud():\n",
    "    NODOS_EXPANDIDOS = 0\n",
    "    HISTORIA = [] # Guardaremos aquí disposiciones ya exploradas de la TORRE, el orden de la mesa es indiferente\n",
    "\n",
    "    def iterar(estados):\n",
    "        print(\"Evaluando un nivel\")\n",
    "        nonlocal NODOS_EXPANDIDOS, HISTORIA\n",
    "        next_estados = []\n",
    "\n",
    "        for estado in estados:\n",
    "            if estado.torre in HISTORIA:\n",
    "                #do nothing\n",
    "                print(\"Nodo repetido\")\n",
    "            elif estado.isMeta():\n",
    "                return estado;\n",
    "            else:\n",
    "                HISTORIA.append(estado.torre)\n",
    "                next_estados += estado.expandirNodo()\n",
    "        NODOS_EXPANDIDOS += len(next_estados)\n",
    "        return iterar(next_estados)\n",
    "\n",
    "    start_time = time.time()\n",
    "\n",
    "    solucion = iterar([Estado(TORRE, MESA, [])])\n",
    "\n",
    "    end_time = time.time()\n",
    "    print(\"SOLUCIÓN {} ENCONTRADA EN {}s.\\nNODOS EXPANDIDOS: {}\\nOrden de movimientos: {}\".format(solucion.torre, end_time - start_time, NODOS_EXPANDIDOS, solucion.solucion))\n",
    "    acciones = solucion.solucion\n",
    "    estado = Estado(TORRE, MESA, [])\n",
    "    estado.print()\n",
    "    for accion in acciones:\n",
    "        estado.moverCaja(accion, True)\n",
    "        estado.print()\n",
    "\n",
    "        \n",
    "# Nota, sin revisar loops, expande 98 nodos\n",
    "# Evitando loops: 49"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {
    "collapsed": false
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Evaluando un nivel\n",
      "Evaluando un nivel\n",
      "Evaluando un nivel\n",
      "Nodo repetido\n",
      "Nodo repetido\n",
      "Nodo repetido\n",
      "Evaluando un nivel\n",
      "Nodo repetido\n",
      "Nodo repetido\n",
      "Nodo repetido\n",
      "Nodo repetido\n",
      "Nodo repetido\n",
      "Evaluando un nivel\n",
      "SOLUCIÓN ['E', 'D', 'C', 'B', 'A'] ENCONTRADA EN 0.004745006561279297s.\n",
      "NODOS EXPANDIDOS: 49\n",
      "Orden de movimientos: ['A', 'C', 'B', 'A']\n",
      "--Estado---------\n",
      "TORRE :: ['E', 'D', 'A']\n",
      "MESA  :: ['C', 'B']\n",
      "Mover a caja [A] de la torre a la mesa\n",
      "--Estado---------\n",
      "TORRE :: ['E', 'D']\n",
      "MESA  :: ['C', 'B', 'A']\n",
      "Mover la caja [C] de la mesa a la torre\n",
      "--Estado---------\n",
      "TORRE :: ['E', 'D', 'C']\n",
      "MESA  :: ['B', 'A']\n",
      "Mover la caja [B] de la mesa a la torre\n",
      "--Estado---------\n",
      "TORRE :: ['E', 'D', 'C', 'B']\n",
      "MESA  :: ['A']\n",
      "Mover la caja [A] de la mesa a la torre\n",
      "--Estado---------\n",
      "TORRE :: ['E', 'D', 'C', 'B', 'A']\n",
      "MESA  :: []\n"
     ]
    }
   ],
   "source": [
    "busquedaAmplitud()\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "collapsed": false,
    "scrolled": true
   },
   "source": [
    "#### Búsqueda con A*"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {
    "collapsed": false,
    "scrolled": true
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Iterando\n",
      "1\n",
      "Estado actual ['E', 'D', 'A'] con coste 4\n",
      "Iterando\n",
      "4\n",
      "Estado actual ['E', 'D'] con coste 3\n",
      "Iterando\n",
      "8\n",
      "Estado actual ['E', 'D', 'C'] con coste 2\n",
      "Iterando\n",
      "11\n",
      "Estado actual ['E', 'D', 'C', 'B'] con coste 1\n",
      "Iterando\n",
      "13\n",
      "Estado actual ['E', 'D', 'C', 'B', 'A'] con coste 0\n",
      "SOLUCIÖN ['E', 'D', 'C', 'B', 'A'] ENCONTRADA EN 0.004506111145019531s.\n",
      "NODOS EXPANDIDOS: 13\n",
      "Orden de movimientos: ['A', 'C', 'B', 'A']\n",
      "--Estado---------\n",
      "TORRE :: ['E', 'D', 'A']\n",
      "MESA  :: ['C', 'B']\n",
      "Mover a caja [A] de la torre a la mesa\n",
      "--Estado---------\n",
      "TORRE :: ['E', 'D']\n",
      "MESA  :: ['C', 'B', 'A']\n",
      "Mover la caja [C] de la mesa a la torre\n",
      "--Estado---------\n",
      "TORRE :: ['E', 'D', 'C']\n",
      "MESA  :: ['B', 'A']\n",
      "Mover la caja [B] de la mesa a la torre\n",
      "--Estado---------\n",
      "TORRE :: ['E', 'D', 'C', 'B']\n",
      "MESA  :: ['A']\n",
      "Mover la caja [A] de la mesa a la torre\n",
      "--Estado---------\n",
      "TORRE :: ['E', 'D', 'C', 'B', 'A']\n",
      "MESA  :: []\n"
     ]
    }
   ],
   "source": [
    "# Algoritmo A*\n",
    "\n",
    "def heuristica(estado):\n",
    "    # Por cada caja en su sitio 0 puntos\n",
    "    # por cada caja faltante 1 punto\n",
    "    # Desde la primera caja mal situada: 2 x cajas por encima, ella incluida\n",
    "    #  Para penalizar el caso en que cajas por encima de ella estén bien\n",
    "    valor = len(estado.mesa) # Sumo cajas faltantes\n",
    "    for idx,caja in enumerate(estado.torre):\n",
    "        # Busco la primera caja mal situada\n",
    "        # ej: en [e,d,a,b] sería a\n",
    "        # la penalización será de dos puntos por todas las cajas de a en adelante, incluso si b está en la posición que le toca\n",
    "        # pues aunque esté bien he de quitarla y volverla a poner\n",
    "        # penalización sería: 4 en este caso, programáticamente: 2*(long_maxima_torre - idx_primera_caja_max - cajas_en_mesa)\n",
    "        # 2*(5 - 2 - 1) = 4\n",
    "        if Estado.META[idx] != caja : \n",
    "            valor += 2 * (5 - idx - len(estado.mesa)) \n",
    "    return valor\n",
    "\n",
    "def busquedaA():\n",
    "    NODOS_EXPANDIDOS = 0\n",
    "    \n",
    "    def iterar(estados_exp):\n",
    "        nonlocal NODOS_EXPANDIDOS\n",
    "        print(\"Iterando\")\n",
    "        print(len(estados_exp))\n",
    "        # Busco el estado con menor valor de función heurística\n",
    "        minimo = Estado([],[],[]) # Inicializado con un estado vacío (coste es infinito)\n",
    "        for estado in estados_exp:\n",
    "            if estado.coste < minimo.coste:\n",
    "                minimo = estado\n",
    "        estado_actual = minimo\n",
    "        \n",
    "        print(\"Estado actual {} con coste {}\".format(estado_actual.torre, estado_actual.coste))\n",
    "        #estados_exp.remove(estado_actual)\n",
    "        \n",
    "        # Comprobamos si estamos en el estado meta\n",
    "        if estado_actual.isMeta():\n",
    "            return estado_actual\n",
    "        \n",
    "        #Si no,  Una vez tengo el nodo mínimo lo expando y aplico la funcion heurística\n",
    "        \n",
    "        #nuevos_nodos = map(lambda e: (e, heuristica(e)),estado.expandirNodo())\n",
    "        nuevos_nodos = [e.set_coste(heuristica(e)) for e in estado_actual.expandirNodo()]\n",
    "        \n",
    "        NODOS_EXPANDIDOS += len(nuevos_nodos)\n",
    "        #print(nuevos_nodos + estados)\n",
    "        # e itero\n",
    "        return iterar(nuevos_nodos + estados_exp)\n",
    "\n",
    "    start_time = time.time()\n",
    "    estado_inicial = Estado(TORRE, MESA, [])\n",
    "    estado_inicial.set_coste(heuristica(estado_inicial))\n",
    "    NODOS_EXPANDIDOS+=1\n",
    "    solucion = iterar([estado_inicial])\n",
    "\n",
    "    end_time = time.time()\n",
    "    print(\"SOLUCIÓN {} ENCONTRADA EN {}s.\\nNODOS EXPANDIDOS: {}\\nOrden de movimientos: {}\".format(solucion.torre, end_time - start_time, NODOS_EXPANDIDOS, solucion.solucion))\n",
    "    acciones = solucion.solucion\n",
    "    estado = Estado(TORRE, MESA, [])\n",
    "    estado.print()\n",
    "    for accion in acciones:\n",
    "        estado.moverCaja(accion, True)\n",
    "        estado.print()\n",
    "        \n",
    "        \n",
    "MESA = [\"C\", \"B\"]\n",
    "TORRE = [\"E\", \"D\", \"A\"]\n",
    "busquedaA()\n",
    "\n",
    "# Nota, sin controlar loops expande 13 nodos\n",
    "#"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "collapsed": false
   },
   "source": [
    "## Conclusión"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "collapsed": false
   },
   "source": [
    "Como conclusión, se ha podido observar que el algoritmo heurístico de A* necesita expandir un número considerablemente menor de nodos que el algoritmo en amplitud. En concreto, el algoritmo en amplitud necesita expandir 98 nodods para alacanzar la meta (49 si se restingen los nodos hacia estados *padre*), frente a los 13 nodos expandidos con la heurística del algoritmo A*."
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (Ubuntu Linux)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.6.7"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 0
}